/***
 *  $Id$
 **
 *  File: closet-cluster-list.cpp
 *  Created: Mar 23, 2012
 *
 *  Author: Jaroslaw Zola <jaroslaw.zola@gmail.com>
 *
 *  This file is part of CLOSET.
 *
 *  Copyright (c) 2011-2012 Xiao Yang, Jaroslaw Zola, Srinivas Aluru
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 */

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>

#include <inttypes.h>

#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/bzip2.hpp>

#include <jaz/iterator_add.hpp>
#include <jaz/string_add.hpp>

#include "SequenceMap.hpp"
#include "ioformat.hpp"


namespace fs = boost::filesystem;


int main(int argc, char* argv[]) {
    if (argc != 4) {
	std::cout << "Usage: closet-cluster-list <map> <cluster> <output>" << std::endl;
	return -1;
    }

    std::string map = argv[1];
    std::string dir = argv[2];
    std::string output = argv[3];

    // get sequence map
    SequenceMap sm;
    SequenceMap::const_iterator miter;

    if (sm.read(map) == false) {
	std::cerr << error << "unable to read sequence map" << std::endl;
	return -1;
    }

    // read cluster files
    fs::path p(dir);
    if (fs::exists(p) == false) return false;

    std::vector<fs::path> files;

    if (fs::is_directory(p) == true) std::copy(fs::directory_iterator(p), fs::directory_iterator(), std::back_inserter(files));
    else files.push_back(p);

    std::istringstream is;
    unsigned int files_sz = files.size();

    std::ofstream of(output.c_str());

    for (unsigned int i = 0; i < files_sz; ++i) {
	if (fs::is_regular_file(files[i]) == true) {
	    std::istream* fp = 0;

	    std::ifstream f;
	    boost::iostreams::filtering_istream fi;

	    if (files[i].extension() == ".bz2") {
		f.open(files[i].string().c_str(), std::ios_base::binary);
		if (!f) {
		    std::cout << "ignoring " << files[i].string() << std::endl;
		    continue;
		}
		fi.push(boost::iostreams::bzip2_decompressor());
		fi.push(f);
		fp = &fi;
	    } else {
		f.open(files[i].string().c_str());
		if (!f) {
		    std::cout << "ignoring " << files[i].string() << std::endl;
		    continue;
		}
		fp = &f;
	    }

	    std::cout << "processing " << files[i].string() << std::endl;

	    jaz::getline_iterator<> li(*fp), end;

	    for (; li != end; ++li) {
		is.clear();
		is.str(*li);

		uint64_t id;
		is >> id;

		std::string s;
		std::getline(is, s);

		if (s.empty() == false) s.erase(s.begin());
		if (s[s.size() - 1] == ' ') s.resize(s.size() - 1);

		std::replace(s.begin(), s.end(), ':', ' ');

		std::vector<std::string> v;
		jaz::split(' ', s, std::back_inserter(v));

		std::vector<unsigned int> nodes(v.size());
		for (unsigned int j = 0; j < v.size(); ++j) {
		    nodes[j] = boost::lexical_cast<unsigned int>(v[j]);
		}

		std::sort(nodes.begin(), nodes.end());
		nodes.erase(std::unique(nodes.begin(), nodes.end()), nodes.end());

		of << ">" << id << std::endl;

		for (unsigned int j = 0; j < nodes.size(); ++j) {
		    miter = sm.find(nodes[j]);
		    if (miter != sm.end()) {
			std::copy(miter->name.begin(), miter->name.end(), std::ostream_iterator<std::string>(of, "\n"));
		    }
		}

		of << std::endl;
	    } // for li

	} // if
    } // for i

    of.close();

    std::cout << "done" << std::endl;

    return 0;
} // main
