/***
 *  $Id$
 **
 *  File: mr-clique-task0.cpp
 *  Created: Sep 27, 2010
 *
 *  Author: Jaroslaw Zola <jaroslaw.zola@gmail.com>
 *          Xiao Yang <eks.yang@gmail.com>
 *
 *  This file is part of CLOSET.
 *
 *  Copyright (c) 2011-2012 Xiao Yang, Jaroslaw Zola, Srinivas Aluru
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 */

#include <set>
#include <stdexcept>
#include <string>
#include <vector>

#include <hdfs.h>
#include <inttypes.h>

#include <hadoop/Pipes.hh>
#include <hadoop/TemplateFactory.hh>

#include <jaz/functional_add.hpp>
#include <jaz/string_add.hpp>

#include "default.h"


/** In theory entire stage0 could be implemented
    as a mapper in stage1. However, because we are
    creating and storing mset_ we have to split
    stage0 and stage1 into separate tasks!
*/

class CliqueMap : public HadoopPipes::Mapper {
public:
    CliqueMap(HadoopPipes::TaskContext& context) {
	const HadoopPipes::JobConf* conf = context.getJobConf();

	irun_ = false;
	simstep_ = CLOSET_SIMSTEP;

	if (conf->hasKey("closet.sim.init.run")) {
	    if (conf->get("closet.sim.init.run") == "1") irun_ = true;
	}

	if (conf->hasKey("closet.sim.step")) {
	    if (st_(conf->get("closet.sim.step"), simstep_) == false) {
		throw std::runtime_error("incorrect sim.step");
	    }
	}

	if ((simstep_ < CLOSET_SIMSTEP_MIN) || (simstep_ > CLOSET_SIMSTEP_MAX)) {
	    throw std::runtime_error("incorrect sim.step");
        }

	if (conf->hasKey("closet.sim.current") == false) {
	    throw std::runtime_error("no sim.current");
	}

	if (st_(conf->get("closet.sim.current"), cursim_) == false) {
	    throw std::runtime_error("incorrect sim.current");
	}
    } // CliqueMap


    void map(HadoopPipes::MapContext& context) {
	std::string key[2];
	jaz::split(' ', context.getInputKey(), key);

	unsigned int sim;
	st_(context.getInputValue(), sim);

	if (irun_ == true) {
	    if (cursim_ < sim) context.emit(key[0], key[1]);
	}
	else {
	    if ((cursim_ < sim) && (sim <= (cursim_ + simstep_))) context.emit(key[0], key[1]);
	}
    } // map


private:
    jaz::string_to st_;

    bool irun_;
    unsigned int cursim_;
    unsigned int simstep_;

}; // class CliqueMap


class CliqueReduce : public HadoopPipes::Reducer {
public:
    CliqueReduce(HadoopPipes::TaskContext& context) {
	const HadoopPipes::JobConf* conf = context.getJobConf();

	// id_ = conf->get("mapreduce.task.attempt.id");
	// dir_ = conf->get("mapreduce.output.fileoutputformat.outputdir");

	id_ = conf->get("mapred.task.id");
	dir_ = conf->get("mapred.work.output.dir");
    } // CliqueReduce


    ~CliqueReduce() {
	if (mset_.empty() == false) {
	    // context.setStatus("writing " + ts_(mset_.size()));

	    std::set<std::string>::iterator iter(mset_.begin());
	    std::string s = *(iter++);

	    for (; iter != mset_.end(); ++iter) s += '\n' + *iter;

	    std::string path = dir_ + "/" + "closet_" + id_ + ".txt";

	    hdfsFS fs = hdfsConnect("default", 0);
	    if (fs == 0) throw std::runtime_error("unable to connect to HDFS");

	    hdfsFile wf = hdfsOpenFile(fs, path.c_str(), O_WRONLY | O_CREAT, 0, 0, 0);
	    if (wf == 0) throw std::runtime_error("unable to create " + path);

	    // context.setStatus("sending to HDFS " + ts_(s.size()));

	    hdfsWrite(fs, wf, (void*)s.c_str(), s.size());
	    if (hdfsFlush(fs, wf)) throw std::runtime_error("unable to flush HDFS");

	    hdfsCloseFile(fs, wf);

	    // context.setStatus("changing mode");

	    // 436 = (644)octal
	    hdfsChmod(fs, path.c_str(), 436);
	    hdfsDisconnect(fs);
	} // if mset_.empty()
    } // ~CliqueReduce


    void reduce(HadoopPipes::ReduceContext& context) {
	std::string key = context.getInputKey();
	mset_.insert(key);

	uint64_t ns[2];
	const char* nsc = reinterpret_cast<const char*>(ns);

	st_(key, ns[0]);

	std::string val;
	std::string e;

	while (context.nextValue()) {
	    val = context.getInputValue();
	    mset_.insert(val);
	    st_(val, ns[1]);
	    e = std::min(key, val) + ':' + std::max(key, val);
	    context.emit(ts_(h_(nsc, 2 * sizeof(uint64_t))), e);
	} // while
    } // reduce


private:
    typedef jaz::murmur264 Hash;

    // mset_ is the list of nodes that should
    // be considered during clustering
    // we will store it in temp file to
    // use in later stages
    std::set<std::string> mset_;

    std::string id_;
    std::string dir_;

    jaz::string_to st_;
    jaz::to_string ts_;

    Hash h_;

}; // class CliqueReduce


int main(int argc, char* argv[]) {
    // return HadoopPipes::runTask(HadoopPipes::TemplateFactory<CliqueMap, CliqueReduce, Partition>());
    return HadoopPipes::runTask(HadoopPipes::TemplateFactory<CliqueMap, CliqueReduce>());
} // main
