/***
 *  $Id$
 **
 *  File: mr-clique-task1.cpp
 *  Created: Sep 27, 2010
 *
 *  Author: Jaroslaw Zola <jaroslaw.zola@gmail.com>
 *          Xiao Yang <eks.yang@gmail.com>
 *
 *  This file is part of CLOSET.
 *
 *  Copyright (c) 2011-2012 Xiao Yang, Jaroslaw Zola, Srinivas Aluru
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 */

#include <algorithm>
#include <set>
#include <stdexcept>
#include <vector>

#include <hdfs.h>
#include <inttypes.h>
#include <math.h>

#include <hadoop/Pipes.hh>
#include <hadoop/TemplateFactory.hh>

#include <jaz/functional_add.hpp>
#include <jaz/plain_array.hpp>
#include <jaz/string_add.hpp>

#include "default.h"
#include "mr-common.hpp"


class CliqueMap : public HadoopPipes::Mapper {
public:
    CliqueMap(HadoopPipes::TaskContext& context) {
	const HadoopPipes::JobConf* conf = context.getJobConf();

	std::string temp_dir;

	if (conf->hasKey("closet.temp.dir") == false) {
	    throw std::runtime_error("no temp.dir");
	} else temp_dir = conf->get("closet.temp.dir");

	// connect to HDFS
	hdfsFS fs = hdfsConnect("default", 0);
	if (fs == 0) throw std::runtime_error("unable to connect to HDFS");

	// read temp
	int res = hdfsExists(fs, temp_dir.c_str());
	if (res == 0) {
	    int dirnum = 0;
	    hdfsFileInfo* info = hdfsListDirectory(fs, temp_dir.c_str(), &dirnum);

	    if (info == 0) dirnum = 0;

	    // create set of involved nodes
	    for (int i = 0; i < dirnum; ++i) {
		jaz::plain_array<char> buf(info[i].mSize + 1, '\0');
		if (read_hdfs_file(fs, info[i], buf.begin()) == false) {
		    throw std::runtime_error("unable to read " + std::string(info[i].mName));
		}

		std::istringstream is;
		is.rdbuf()->pubsetbuf(buf.begin(), info[i].mSize + 1);

		while (is.eof() == false) {
		    std::string s;
		    is >> s;
		    mset_.insert(s);
		}
	    } // for i

	    hdfsFreeFileInfo(info, dirnum);
	} // if res

	hdfsDisconnect(fs);
    } // CliqueMap


    void map(HadoopPipes::MapContext& context) {
	std::string val = context.getInputValue();

	std::vector<std::string> nd;
	cluster2nodes(val, std::back_inserter(nd));

	// here we check if a cluster participated in merging
	// in the previous iteration, if it was it should be in mset
	bool mp = false;

	for (unsigned int i = 0; i < nd.size(); ++i) {
	    if (mset_.find(nd[i]) != mset_.end()) {
		context.emit(nd[i], val);
		mp = true;
	    }
	} // for

	// if not we just emit it to make sure
	// it will be kept in the output
	if (mp == false) context.emit(nd[0], val);
    } // map


private:
    std::set<std::string> mset_;

}; // class CliqueMap


class CliqueReduce : public HadoopPipes::Reducer {
public:
    CliqueReduce(HadoopPipes::TaskContext& context) {
	const HadoopPipes::JobConf* conf = context.getJobConf();

	// id_ = conf->get("mapreduce.task.attempt.id");
	// dir_ = conf->get("mapreduce.output.fileoutputformat.outputdir");

	id_ = conf->get("mapred.task.id");
	dir_ = conf->get("mapred.work.output.dir");

	maxdeg_ = CLOSET_MDEG;

	if (conf->hasKey("closet.mdeg")) {
	    if (st_(conf->get("closet.mdeg"), maxdeg_) == false) {
		throw std::runtime_error("incorrect mdeg");
	    }
	}

	if (maxdeg_ < CLOSET_MDEG_MIN) throw std::runtime_error("incorrect mdeg");

	gamma_ = CLOSET_GAMMA;

	if (conf->hasKey("closet.gamma")) {
	    if (st_(conf->get("closet.gamma"), gamma_) == false) {
		throw std::runtime_error("incorrect gamma");
	    }
	}

	if ((gamma_ < CLOSET_GAMMA_MIN) || (gamma_ > CLOSET_GAMMA_MAX)) {
	    throw std::runtime_error("incorrect gamma");
	}
    } // CliqueReduce


    ~CliqueReduce() {
	if (mset_.empty() == false) {
	    // context.setStatus("writing " + ts_(mset_.size()));

	    std::set<std::string>::iterator iter(mset_.begin());
	    std::string s = *(iter++);

	    for (; iter != mset_.end(); ++iter) s += '\n' + *iter;

	    std::string path = dir_ + "/" + "closet_" + id_ + ".txt";

	    hdfsFS fs = hdfsConnect("default", 0);
	    if (fs == 0) throw std::runtime_error("unable to connect to HDFS");

	    hdfsFile wf = hdfsOpenFile(fs, path.c_str(), O_WRONLY | O_CREAT, 0, 0, 0);
	    if (wf == 0) throw std::runtime_error("unable to create " + path);

	    // context.setStatus("sending to HDFS " + ts_(s.size()));

	    hdfsWrite(fs, wf, (void*)s.c_str(), s.size());
	    if (hdfsFlush(fs, wf)) throw std::runtime_error("unable to flush HDFS");

	    hdfsCloseFile(fs, wf);

	    // context.setStatus("changing mode");

	    // 436 = (644)octal
	    hdfsChmod(fs, path.c_str(), 436);
	    hdfsDisconnect(fs);
	} // if mset_.empty()
    } // ~CliqueReduce


    void reduce(HadoopPipes::ReduceContext& context) {
	HadoopPipes::ReduceContext::Counter* closet_stars = context.getCounter("CLOSET Counters", "Stars rejected");
	HadoopPipes::ReduceContext::Counter* closet_merge = context.getCounter("CLOSET Counters", "Cliques merged");

	// get all clusters into one string
	std::vector<std::string> S;

	while (context.nextValue()) {
	    S.push_back(context.getInputValue());
	    if (S.size() > maxdeg_) {
		context.incrementCounter(closet_stars, 1);
		return;
	    }
	}

	unsigned int l = S.size();

	// get list of nodes and edges for each cluster
	std::vector<std::vector<std::string> > nlst(l);
	std::vector<std::vector<std::string> > elst(l);

	for (unsigned int i = 0; i < l; ++i) {
	    cluster2nodes(S[i], std::back_inserter(nlst[i]));
	    jaz::split(' ', S[i], std::back_inserter(elst[i]));
	    std::sort(elst[i].begin(), elst[i].end());
	} // for i

	// if only one element job is easy
	if (l == 1) {
	    uint64_t nsh = nodes2hash(nlst[0].begin(), nlst[0].end(), h_);
	    std::string ns = ts_(nsh);
	    context.emit(ns, S[0]);
	    context.setStatus("done");
	    return;
	} // if l

	// *** here tricky part begins ***

	// this will be our new input
	// if more than one merging step is needed
	std::vector<std::vector<std::string> > nlst2;
	std::vector<std::vector<std::string> > elst2;

	nlst2.reserve(l);
	elst2.reserve(l);

	unsigned int nt = 0;

	while (true) {
	    nt++;

	    // flag to mark clusters that are connected to form
	    // gamma clique
	    l = nlst.size();
	    std::vector<bool> flg(l, false);

	    context.setStatus("trying to form cliques: " + ts_(nt));

	    // check all pairs of clusters if they form gamma clique
	    for (unsigned int i = 0; i < l - 1; ++i) {
		for (unsigned int j = i + 1; j < l; ++j) {

		    // make cluster
		    std::vector<std::string> nunion;
		    std::vector<std::string> eunion;

		    nunion.reserve(nlst[i].size() + nlst[j].size());
		    eunion.reserve(elst[i].size() + elst[j].size());

		    std::set_union(nlst[i].begin(), nlst[i].end(),
				   nlst[j].begin(), nlst[j].end(),
				   std::back_inserter(nunion));

		    std::set_union(elst[i].begin(), elst[i].end(),
				   elst[j].begin(), elst[j].end(),
				   std::back_inserter(eunion));

		    unsigned int nusz = nunion.size();
		    bool merge = true;

		    // // check cluster (based on nodes)
		    // unsigned int d = 1 + ((gamma_ * nusz) / 100);
		    // unsigned int  d = static_cast<unsigned int>(log2(1 + ((1.0 * gamma_ * nusz) / 100)));

		    // std::vector<std::string> e;
		    // for (unsigned int k = 0; k < eunion.size(); ++k) {
		    // 	jaz::split(':', eunion[k], std::back_inserter(e));
		    // }

		    // for (unsigned int k = 0; k < nusz; ++k) {
		    // 	if (std::count(e.begin(), e.end(), nunion[k]) < d) {
		    // 	    merge = false;
		    // 	    break;
		    // 	}
		    // } // for k

		    // check cluster (based on edges)
		    unsigned int d = nusz * nusz - nusz;
		    d = (200 * eunion.size()) / d;
		    // first density

		    if (d >= gamma_) {
			// if we have more than three nodes check for tips
			if (nusz > 3) {
			    std::vector<std::string> e;
			    for (unsigned int k = 0; k < eunion.size(); ++k) {
				jaz::split(':', eunion[k], std::back_inserter(e));
			    }
			    for (unsigned int k = 0; k < nusz; ++k) {
				if (std::count(e.begin(), e.end(), nunion[k]) < 2) {
				    merge = false;
				    break;
				}
			    } // for k
			} // if nusz
		    } else merge = false;

		    // perform merging
		    if (merge == true) {
			mset_.insert(nunion.begin(), nunion.end());
			nlst2.push_back(nunion);
			elst2.push_back(eunion);
			flg[i] = flg[j] = true;
		    }

		} // for j
		context.progress();
	    } // for i

	    context.setStatus("done forming cliques: " + ts_(nt));

	    bool emit = true;

	    for (unsigned int i = 0; i < l; ++i) {
		// if a cluster did not form gamma clique
		// in this iteration it may form it later
		// or if no single pair formed a cluster
		// we are done and should emit
		if (flg[i] == false) {
		    nlst2.push_back(nlst[i]);
		    elst2.push_back(elst[i]);
		} else emit = false;
	    } // for i

	    // if no pair formed a clique we are done
	    if (emit == true) {
		l = nlst2.size();
		for (unsigned int i = 0; i < l; ++i) {
		    uint64_t nsh = nodes2hash(nlst2[i].begin(), nlst2[i].end(), h_);
		    std::string ns = ts_(nsh);
		    std::string es = jaz::join(' ', elst2[i].begin(), elst2[i].end());
		    context.emit(ns, es);
		}
		break;
	    } else {
		// now we have to merge all cliques
		// that got the same list of nodes
		context.setStatus("merging cliques: " + ts_(nt));

		std::vector<NE> une(nlst2.size());

		for (unsigned int i = 0; i < nlst2.size(); ++i) {
		    une[i].pos = i;
		    une[i].nds = jaz::join(' ', nlst2[i].begin(), nlst2[i].end());;
		}

		std::sort(une.begin(), une.end());

		// here we prepare input for the next iteration
		// of clusters merging
		nlst.clear();
		elst.clear();

		unsigned int pos = 0;

		// combine cliques with the same set of nodes
		for (unsigned int j = 1; j < nlst2.size(); ++j) {
		    if (une[pos].nds != une[j].nds) {
			context.incrementCounter(closet_merge, (j - pos - 1));

			nlst.resize(nlst.size() + 1);
			nlst[nlst.size() - 1] = nlst2[une[pos].pos];
			elst.resize(elst.size() + 1);

			for (unsigned int k = pos; k < j; ++k) {
			    std::copy(elst2[une[k].pos].begin(), elst2[une[k].pos].end(), std::back_inserter(elst[elst.size() - 1]));

			}

			std::sort(elst[elst.size() - 1].begin(), elst[elst.size() - 1].end());
			std::vector<std::string>::iterator it = std::unique(elst[elst.size() - 1].begin(), elst[elst.size() - 1].end());
			elst[elst.size() - 1].resize(it - elst[elst.size() - 1].begin());

			pos = j;
		    } // if
		} // for

		// take care of the last batch
		context.incrementCounter(closet_merge, (nlst2.size() - pos - 1));

		nlst.resize(nlst.size() + 1);
		nlst[nlst.size() - 1] = nlst2[une[pos].pos];
		elst.resize(elst.size() + 1);

		for (unsigned int k = pos; k < nlst2.size(); ++k) {
		    std::copy(elst2[une[k].pos].begin(), elst2[une[k].pos].end(), std::back_inserter(elst[elst.size() - 1]));
		}

		unsigned int sz = elst.size();
		std::sort(elst[sz - 1].begin(), elst[sz - 1].end());
		std::vector<std::string>::iterator it = std::unique(elst[sz - 1].begin(), elst[sz - 1].end());
		elst[sz - 1].resize(it - elst[sz - 1].begin());

		// now(?) we can clear :-)
		nlst2.clear();
		elst2.clear();

		// if we exceeded maxdeg_ we are done
		if (nlst.size() > maxdeg_) {
		     l = nlst.size();
		     for (unsigned int i = 0; i < l; ++i) {
		 	uint64_t nsh = nodes2hash(nlst[i].begin(), nlst[i].end(), h_);
		 	std::string ns = ts_(nsh);
		 	std::string es = jaz::join(' ', elst[i].begin(), elst[i].end());
		 	context.emit(ns, es);
		     }
		     // mset_.erase(context.getInputKey()); // not sure about this really
		     break;
		}
	    } // if else emit
	} // while (true)

	context.setStatus("done");
    } // reduce


private:
    struct NE {
	unsigned int pos;
	std::string nds;
	bool operator<(const NE& rhs) const { return nds < rhs.nds; }
    }; // struct NE

    typedef jaz::murmur264 Hash;

    std::set<std::string> mset_;

    jaz::string_to st_;
    jaz::to_string ts_;

    Hash h_;

    std::string id_;
    std::string dir_;

    unsigned int maxdeg_;
    unsigned int gamma_;

}; // class CliqueReduce


int main(int argc, char* argv[]) {
    // return HadoopPipes::runTask(HadoopPipes::TemplateFactory<CliqueMap, CliqueReduce, CliquePartition>());
    return HadoopPipes::runTask(HadoopPipes::TemplateFactory<CliqueMap, CliqueReduce>());
} // main
