/***
 *  $Id$
 **
 *  File: mr-sketch-task0.cpp
 *  Created: Sep 07, 2010
 *
 *  Author: Jaroslaw Zola <jaroslaw.zola@gmail.com>
 *          Xiao Yang <eks.yang@gmail.com>
 *
 *  This file is part of CLOSET.
 *
 *  Copyright (c) 2011-2012 Xiao Yang, Jaroslaw Zola, Srinivas Aluru
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdexcept>

#include <hdfs.h>
#include <inttypes.h>

#include <boost/lexical_cast.hpp>

#include <hadoop/Pipes.hh>
#include <hadoop/TemplateFactory.hh>

#include <jaz/functional_add.hpp>

#include "default.h"

#include "SequenceCodec.hpp"
#include "nascii85.hpp"


class SketchMap : public HadoopPipes::Mapper {
public:
    SketchMap(HadoopPipes::TaskContext& context) {
	const HadoopPipes::JobConf* conf = context.getJobConf();

	kvalue_ = CLOSET_KVALUE;

	if (conf->hasKey("closet.kvalue")) {
	    kvalue_ = boost::lexical_cast<unsigned int>(conf->get("closet.kvalue"));
	}

	if ((kvalue_ < CLOSET_KVALUE_MIN) || (kvalue_ > CLOSET_KVALUE_MAX)) {
	    throw std::runtime_error("incorrect kvalue");
	}

	mvalue_ = 1 / CLOSET_MVALUE;

	if (conf->hasKey("closet.mvalue")) {
	    double m = boost::lexical_cast<double>(conf->get("closet.mvalue"));
	    if ((m <= CLOSET_MVALUE_MIN) || (m >= CLOSET_MVALUE_MAX)) {
		throw std::runtime_error("incorrect mvalue");
	    }
	    mvalue_ = 1 / m;
	}

	mcurrent_ = 0;

	if (conf->hasKey("closet.mvalue.current") == false) {
	    throw std::runtime_error("no mvalue.current");
	}

	mcurrent_ = boost::lexical_cast<unsigned int>(conf->get("closet.mvalue.current"));

	if (mcurrent_ >= mvalue_) {
	    throw std::runtime_error("incorrect mvalue.current");
	}
    } // SketchMap


    void map(HadoopPipes::MapContext& context) {
	std::string sid = context.getInputKey();
	std::string s = SQ_.decode(context.getInputValue());

	unsigned int l = s.length();

	if ((kvalue_ <= l) && (mvalue_ <= l)) {
	    for (unsigned int i = 0; i < l - kvalue_ + 1; ++i) {
		uint64_t hv = h_(SQ_.code(std::string(s.begin() + i, s.begin() + i + kvalue_)));
		if (hv % mvalue_ == mcurrent_) context.emit(nascii85_encode(hv), sid);
	    }
	} // if
    } // map


private:
    typedef jaz::murmur264 Hash;

    SequenceCodec SQ_;
    Hash h_;

    unsigned int kvalue_;
    unsigned int mvalue_;
    unsigned int mcurrent_;

}; // class SketchMap


class SketchReduce : public HadoopPipes::Reducer {
public:
    SketchReduce(HadoopPipes::TaskContext& context) {
	const HadoopPipes::JobConf* conf = context.getJobConf();

	scount_ = CLOSET_SCOUNT;

	if (conf->hasKey("closet.scount")) {
	    scount_ = boost::lexical_cast<unsigned int>(conf->get("closet.scount"));
	}

	if (scount_ < CLOSET_SCOUNT_MIN) {
	    throw std::runtime_error("incorrect scount");
	}

	// create file to store data
	std::string id = conf->get("mapred.task.id");
	std::string dir = conf->get("mapred.work.output.dir");

	path_ = dir + "/" + "closet_" + id + ".txt";

	fs_ = hdfsConnect("default", 0);
	if (fs_ == 0) throw std::runtime_error("unable to connect to HDFS");

	wf_ = hdfsOpenFile(fs_, path_.c_str(), O_WRONLY | O_CREAT, 0, 0, 0);
	if (wf_ == 0) throw std::runtime_error("unable to create " + path_);

	// 436 = (644)octal
	hdfsChmod(fs_, path_.c_str(), 436);
    } // SketchReduce

    ~SketchReduce() {
	hdfsCloseFile(fs_, wf_);
	hdfsDisconnect(fs_);
    } // ~SketchReduce

    void reduce(HadoopPipes::ReduceContext& context) {
	std::string S = "";
	unsigned int l = 0;

	std::string key = context.getInputKey();

	// put first scount_ sequence ids into one string
	while (context.nextValue()) {
	    S += context.getInputValue();
	    ++l;
	}

	if (l == 1) return;

	// if number of sequence ids is less then scount_ we emit them
	if (l < scount_) context.emit(nascii85_encode(l), S);
	else {
	    // if there are more than scount_ we store them separately
	    S += "\n";
	    hdfsWrite(fs_, wf_, (void*)S.c_str(), S.size());
	    if (hdfsFlush(fs_, wf_)) throw std::runtime_error("unable to flush HDFS");
	}
    } // reduce


private:
    hdfsFS fs_;
    hdfsFile wf_;

    std::string path_;

    unsigned int scount_;

}; // class SketchReduce


int main(int argc, char* argv[]) {
    // return HadoopPipes::runTask(HadoopPipes::TemplateFactory<SketchMap, SketchReduce, Partition>());
    return HadoopPipes::runTask(HadoopPipes::TemplateFactory<SketchMap, SketchReduce>());
} // main
