/***
 *  $Id$
 **
 *  File: mr-sketch-task1.cpp
 *  Created: Sep 10, 2010
 *
 *  Author: Jaroslaw Zola <jaroslaw.zola@gmail.com>
 *          Xiao Yang <eks.yang@gmail.com>
 *
 *  This file is part of CLOSET.
 *
 *  Copyright (c) 2011-2012 Xiao Yang, Jaroslaw Zola, Srinivas Aluru
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdexcept>
#include <vector>

#include <hdfs.h>
#include <inttypes.h>

#include <boost/lexical_cast.hpp>
#include <boost/tuple/tuple.hpp>

#include <hadoop/Pipes.hh>
#include <hadoop/TemplateFactory.hh>

#include <jaz/plain_array.hpp>

#include "SequenceIndex.hpp"
#include "mr-common.hpp"
#include "nascii85.hpp"

#include "default.h"


class SketchMap : public HadoopPipes::Mapper {
public:
    // !!!!" represents 1 in ascii85
    SketchMap(HadoopPipes::TaskContext& context) : one_("!!!!\"") { }

    void map(HadoopPipes::MapContext& context) {
	std::string S = context.getInputValue();

	unsigned int l = S.size() / 5;
	jaz::plain_array<unsigned int> sid(l);

	nascii85_decode(S, sid.begin(), l);

	for (unsigned int i = 0; i < l - 1; ++i) {
	    for (unsigned int j = i + 1; j < l; ++j) {
		if (sid[i] != sid[j]) {
		    unsigned int id[2];
		    id[0] = sid[i];
		    id[1] = sid[j];
		    if (id[1] < id[0]) std::swap(id[0], id[1]);
		    context.emit(nascii85_encode(id, 2), one_);
		}
	    } // for j
	} // for i
    } // map

private:
    std::string one_;

}; // class SketchMap


class SketchCombine : public HadoopPipes::Reducer {
public:
    SketchCombine(HadoopPipes::TaskContext& context) { }

    void reduce(HadoopPipes::ReduceContext& context) {
	std::string k = context.getInputKey();

	// do basic counting
	unsigned int count = 0;
	while (context.nextValue()) {
	    count += nascii85_decode<unsigned int>(context.getInputValue());
	}

	context.emit(k, nascii85_encode(count));
    } // reduce

}; // class SketchCombine


class SketchReduce : public HadoopPipes::Reducer {
public:
    SketchReduce(HadoopPipes::TaskContext& context) {
	const HadoopPipes::JobConf* conf = context.getJobConf();

	// we need explicit access to logic dir
	if (conf->hasKey("closet.logic.dir") == false) {
	    throw std::runtime_error("no logic.dir");
	} else logic_dir_ = conf->get("closet.logic.dir");

	// and to temp dir
	if (conf->hasKey("closet.temp.dir") == false) {
	    throw std::runtime_error("no temp.dir");
	} else temp_dir_ = conf->get("closet.temp.dir");

	// and to input dir
	if (conf->hasKey("closet.input.dir") == false) {
	    throw std::runtime_error("no input.dir");
	} else input_dir_ = conf->get("closet.input.dir");

	minsim_ = CLOSET_SIMMIN;

	if (conf->hasKey("closet.sim.min")) {
	    minsim_ = boost::lexical_cast<unsigned int>(conf->get("closet.sim.min"));
	}

	if ((minsim_ < CLOSET_SIMMIN_MIN) || (minsim_ > CLOSET_SIMMIN_MAX)) {
	    throw std::runtime_error("incorrect sim.min");
        }

	kvalue_ = CLOSET_KVALUE;

	if (conf->hasKey("closet.kvalue")) {
	    kvalue_ = boost::lexical_cast<unsigned int>(conf->get("closet.kvalue"));
	}

	if ((kvalue_ < CLOSET_KVALUE_MIN) || (kvalue_ > CLOSET_KVALUE_MAX)) {
	    throw std::runtime_error("incorrect kvalue");
	}

	mvalue_ = 1 / CLOSET_MVALUE;

	if (conf->hasKey("closet.mvalue")) {
	    double m = boost::lexical_cast<double>(conf->get("closet.mvalue"));
	    if ((m <= CLOSET_MVALUE_MIN) || (m >= CLOSET_MVALUE_MAX)) {
		throw std::runtime_error("incorrect mvalue");
	    }
	    mvalue_ = 1 / m;
	}

	// connect to HDFS
	fs_ = hdfsConnect("default", 0);
	if (fs_ == 0) throw std::runtime_error("unable to connect to HDFS");

	// get temp data
	readTemp_();

	HadoopPipes::TaskContext::Counter* closet_srrem = context.getCounter("CLOSET Counters", "SRrem list size");
	context.incrementCounter(closet_srrem, seq_set_.size());

	// get logic data
	readLogic_();

	hdfsDisconnect(fs_);
    } // SketchReduce


    void reduce(HadoopPipes::ReduceContext& context) {
	std::string k = context.getInputKey();
	unsigned int id[2];

	nascii85_decode(k, id, 2);

	// do basic counting
	unsigned int count = 0;
	while (context.nextValue()) {
	    count += nascii85_decode<unsigned int>(context.getInputValue());
	}

	// update count of high frequency elements
	unsigned int sz = seq_set_.size();
	for (unsigned int i = 0; i < sz; ++i) {
	    if (std::binary_search(seq_set_[i].begin(), seq_set_[i].end(), id[0])
		&& std::binary_search(seq_set_[i].begin(), seq_set_[i].end(), id[1])) {
		count++;
	    }
	} // for i

	// derive similarity
	SequenceIndex id0;
	SequenceIndex id1;

	id0.id = id[0];
	id1.id = id[1];

	std::vector<SequenceIndex>::const_iterator s0, s1, last;

	boost::tie(s0, last) = std::equal_range(idx_set_.begin(), idx_set_.end(), id0);
	boost::tie(s1, last) = std::equal_range(idx_set_.begin(), idx_set_.end(), id1);

	// stage0 should/must guarantee that len is always greater mvalue_
	// in order to calculate sim we should know how many sketches were derived
	// for both sequences (id0 and id1), because we do not know this number
	// instead we are approximating it with the expected number
	unsigned int sim = (100 * count) / std::min(s0->len / mvalue_, s1->len / mvalue_);

	if (sim >= minsim_) context.emit(std::string(k.begin(), k.begin() + 5),
					 std::string(k.begin() + 5, k.begin() + 10));
    } // reduce


private:
    void readTemp_() {
	int res = hdfsExists(fs_, temp_dir_.c_str());
	if (res != 0) throw std::runtime_error("unable to access " + temp_dir_);

	int dirnum = 0;
	hdfsFileInfo* dirinfo = hdfsListDirectory(fs_, temp_dir_.c_str(), &dirnum);
	if (dirinfo == 0) dirnum = 0;

	unsigned int pos = 0;
	for (int i = 0; i < dirnum; ++i) scanTempFile_(dirinfo[i], pos);

	hdfsFreeFileInfo(dirinfo, dirnum);
    } // readTemp_

    void scanTempFile_(const hdfsFileInfo& info, unsigned int& pos) {
	if (info.mKind == 'D') return;

	jaz::plain_array<char> buf(info.mSize + 1, '\0');
	if (read_hdfs_file(fs_, info, buf.begin()) == false) {
	    throw std::runtime_error("unable to read " + std::string(info.mName));
	}

	std::istringstream is;
	is.rdbuf()->pubsetbuf(buf.begin(), info.mSize + 1);

	while (is.eof() == false) {
	    std::string ln = "";
	    std::getline(is, ln);
	    if (ln.size() > 4) {
		seq_set_.resize(pos + 1);
		seq_set_[pos].resize(ln.size() / 5);
		nascii85_decode(ln, seq_set_[pos].begin(), seq_set_[pos].size());
		std::sort(seq_set_[pos].begin(), seq_set_[pos].end());
		pos++;
	    }
	} // while is.eof()
    } // scanTempFile_


    void readLogic_() {
	hdfsFileInfo* info = hdfsGetPathInfo(fs_, (logic_dir_ + "/input.idx").c_str());
	if (info == 0) throw std::runtime_error("unable to check " + logic_dir_ + "/input.idx");

	idx_set_.resize(info->mSize / sizeof(SequenceIndex));
	char* buf = reinterpret_cast<char*>(&idx_set_[0]);

	if (read_hdfs_file(fs_, *info, buf) == false) {
	    throw std::runtime_error("unable to read " + std::string(info->mName));
	}

	std::sort(idx_set_.begin(), idx_set_.end());

	hdfsFreeFileInfo(info, 1);
    } // readLogic_


    jaz::string_to st_;

    hdfsFS fs_;

    std::string logic_dir_;
    std::string temp_dir_;
    std::string input_dir_;

    unsigned int kvalue_;
    unsigned int mvalue_;
    unsigned int minsim_;

    typedef jaz::plain_array<unsigned int> set_type;

    std::vector<set_type> seq_set_;
    std::vector<SequenceIndex> idx_set_;

}; // class SketchReduce


int main(int argc, char* argv[]) {
    return HadoopPipes::runTask(HadoopPipes::TemplateFactory<SketchMap, SketchReduce>());
    // return HadoopPipes::runTask(HadoopPipes::TemplateFactory<SketchMap, SketchReduce, void, SketchCombine>());
} // main
