/***
 *  $Id$
 **
 *  File: mr-sketch-task4.cpp
 *  Created: Sep 18, 2010
 *
 *  Author: Jaroslaw Zola <jaroslaw.zola@gmail.com>
 *          Xiao Yang <eks.yang@gmail.com>
 *
 *  This file is part of CLOSET.
 *
 *  Copyright (c) 2011-2012 Xiao Yang, Jaroslaw Zola, Srinivas Aluru
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 */

#include <algorithm>
#include <stdexcept>
#include <vector>

#include <inttypes.h>

#include <boost/lexical_cast.hpp>

#include <hadoop/Pipes.hh>
#include <hadoop/TemplateFactory.hh>

#include <jaz/algorithm_add.hpp>
#include <jaz/string_add.hpp>

#include "SequenceCodec.hpp"
#include "sequence_compare.hpp"
#include "nascii85.hpp"

#include "default.h"


class SketchMap : public HadoopPipes::Mapper {
public:
    SketchMap(HadoopPipes::TaskContext& context) { }

    void map(HadoopPipes::MapContext& context) {
	std::string k = context.getInputKey();
	std::string S = context.getInputValue();

	std::string val[2];
	jaz::split(' ', S, val);

	unsigned int l = val[1].size() / 5;
	std::string s;

	for (unsigned int i = 0; i < l; ++i) {
	    s = std::string(val[1].begin() + 5 * i, val[1].begin() + 5 * (i + 1));
	    if (k < s) context.emit(k + s, val[0]);
	    else context.emit(s + k, val[0]);
	}
    } // map

}; // class SketchMap


class SketchReduce : public HadoopPipes::Reducer {
public:
    SketchReduce(HadoopPipes::TaskContext& context) {
	const HadoopPipes::JobConf* conf = context.getJobConf();

	validsim_ = CLOSET_SIMVALID;

	if (conf->hasKey("closet.sim.valid")) {
	    validsim_ = boost::lexical_cast<unsigned int>(conf->get("closet.sim.valid"));
	}

	if ((validsim_ < CLOSET_SIMMIN_MIN) || (validsim_ > CLOSET_SIMMIN_MAX)) {
	    throw std::runtime_error("incorrect sim.valid");
        }

	method_ = CLOSET_SIMFUN;

	if (conf->hasKey("closet.sim.fun")) {
	    method_ = conf->get("closet.sim.fun");
	}

	std::vector<std::string> methods;
	jaz::split(' ', std::string(CLOSET_SIMFUN_LIST), std::back_inserter(methods));

	if (std::binary_search(methods.begin(), methods.end(), method_) == false) {
	    throw std::runtime_error("incorrect sim.fun");
	}

	unsigned int kvalue = CLOSET_KVALUE;

	if (conf->hasKey("closet.kvalue")) {
	    kvalue = boost::lexical_cast<unsigned int>(conf->get("closet.kvalue"));
	}

	if ((kvalue < CLOSET_KVALUE_MIN) || (kvalue > CLOSET_KVALUE_MAX)) {
	    throw std::runtime_error("incorrect kvalue");
	}

	aln_ = global_alignment(5, -4, -10, -1);
	jid_ = dna_jaccard_index(kvalue);
    } // SketchReduce


    void reduce(HadoopPipes::ReduceContext& context) {
	unsigned int k[2];
	nascii85_decode(context.getInputKey(), k, 2);

	context.nextValue();
	std::string s0 = SQ_.decode(context.getInputValue());

	context.nextValue();
	std::string s1 = SQ_.decode(context.getInputValue());

	unsigned int sim = compareSequences_(s0, s1);

	if (sim >= validsim_) {
	    if (k[1] < k[0]) std::swap(k[0], k[1]);
	    context.emit(boost::lexical_cast<std::string>(k[0]) + ' ' + boost::lexical_cast<std::string>(k[1]),
			 boost::lexical_cast<std::string>(sim));
	}
    } // reduce


private:
    unsigned int compareSequences_(const std::string& s0, const std::string& s1) {
	unsigned int sim = 0;

	if (method_ == "jaccard") sim = jid_(s0, s1);
	else if (method_ == "alignment") sim = aln_(s0, s1);

	return sim;
    } // compareSequences_


    SequenceCodec SQ_;

    unsigned int validsim_;
    std::string method_;

    global_alignment aln_;
    dna_jaccard_index jid_;

}; // class SketchReduce


int main(int argc, char* argv[]) {
    return HadoopPipes::runTask(HadoopPipes::TemplateFactory<SketchMap, SketchReduce>());
} // main
