/***
 *  $Id$
 **
 *  File: nascii85.hpp
 *  Created: Nov 24, 2011
 *
 *  Author: Jaroslaw Zola <jaroslaw.zola@gmail.com>
 *
 *  This file is part of CLOSET.
 *
 *  Copyright (c) 2011-2012 Xiao Yang, Jaroslaw Zola, Srinivas Aluru
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef NASCII85_HPP
#define NASCII85_HPP

#include <string>
#include <jaz/utility_add.hpp>

/** Simplified implementation of Ascii85 encoding. For our needs we assume
 *  that input consists of integers only. Because not many zeros occurs in
 *  our data no 'z' character is used for compression. Hence, we get very
 *  efficient and easy to use code.
 */

inline void ascii85_encode__(unsigned int in, char* out) {
    unsigned int w;
    unsigned int q;

    q = in / (85 * 85 * 85 * 85);
    out[0] = '!' + q;

    in -= q * (85 * 85 * 85 * 85);
    q = in / (85 * 85 * 85);
    out[1] = '!' + q;

    in -= q * (85 * 85 * 85);
    q = in / (85 * 85);
    out[2] = '!' + q;

    w = (in - q * (85 * 85));
    out[3] = '!' + (w / 85);
    out[4] = '!' + (w % 85);
} // ascii85_encode__

inline unsigned int ascii85_decode__(const char* in) {
    unsigned int w = 0;
    w += (in[0] - '!') * 85 * 85 * 85 * 85;
    w += (in[1] - '!') * 85 * 85 * 85;
    w += (in[2] - '!') * 85 * 85;
    w += (in[3] - '!') * 85;
    w += (in[4] - '!');
    return w;
} // ascii85_decode__


template <typename T> std::string nascii85_encode(const T* t, unsigned int n) {
    unsigned int l = (n * sizeof(T)) >> 2;
    const unsigned int* p = reinterpret_cast<const unsigned int*>(t);

    char* buf = new char[5 * l + 1];
    jaz::scoped_array<char> s(buf);

    for (unsigned int i = 0; i < l; ++i, buf += 5) ascii85_encode__(p[i], buf);
    (*buf) = 0;

    return std::string(s.get());
} // nascii85_encode

template <typename T> inline std::string nascii85_encode(T t) {
    const unsigned char* p = reinterpret_cast<unsigned char*>(&t);
    return nascii85_encode(p, sizeof(T));
} // nascii85_encode


template <typename T> void nascii85_decode(const std::string& s, T* t, unsigned int n) {
    unsigned int l = (n * sizeof(T)) >> 2;
    unsigned int* p = reinterpret_cast<unsigned int*>(t);
    const char* buf = s.c_str();
    for (unsigned int i = 0; i < l; ++i, buf += 5) p[i] = ascii85_decode__(buf);
} // nascii85_decode

template <typename T> inline T nascii85_decode(const std::string& s) {
    T t = 0;
    unsigned int l = sizeof(T) >> 2;
    unsigned int* p = reinterpret_cast<unsigned int*>(&t);
    nascii85_decode(s, p, l);
    return t;
} // nascii85_decode

#endif // NASCII85_HPP
