/***
 *  $Id$
 **
 *  File: closet-fasta-preprocess.cpp
 *  Created: Sep 06, 2010
 *
 *  Author: Jaroslaw Zola <jaroslaw.zola@gmail.com>
 *          Xiao Yang <eks.yang@gmail.com>
 *
 *  This file is part of CLOSET.
 *
 *  Copyright (c) 2011-2012 Xiao Yang, Jaroslaw Zola, Srinivas Aluru
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 */

#include <algorithm>
#include <fstream>
#include <iostream>
#include <vector>

#include <bIO/fasta_file.hpp>

#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/unordered_map.hpp>

#include <jaz/disjoint_set.hpp>
#include <jaz/functional_add.hpp>
#include <jaz/plain_array.hpp>

#include "SequenceCodec.hpp"
#include "SequenceIndex.hpp"
#include "ioformat.hpp"
#include "nascii85.hpp"


namespace fs = boost::filesystem;


typedef bIO::FASTA_input::value_type fasta_type;


struct SDescriptor {
    std::string name;
    unsigned int len;
    uint64_t sh;
}; // SDescriptor


void clean_sequence(std::string& s) {
    unsigned int l = s.size();
    for (unsigned int i = 0; i < l; ++i) {
	s[i] = std::toupper(s[i]);
	if ((s[i] != 'A') && (s[i] != 'C') && (s[i] != 'G') && (s[i] != 'T')) {
	    s[i] = 'T';
	}
    }
} // clean_sequence


uint64_t super_shingle(const std::string& s, unsigned int kmer1, unsigned int kmer2) {
    // stage 1
    unsigned int l = s.size();
    jaz::plain_array<uint64_t> sh1(l - kmer1 + 1, 0);

    SequenceCodec sc;
    jaz::murmur264 hash;

    for (unsigned int i = 0; i < l - kmer1 + 1; ++i) {
	sh1[i] = hash(sc.code(std::string(s.begin() + i, s.begin() + i + kmer1)));
    }

    std::sort(sh1.begin(), sh1.end());

    // stage 2
    l = sh1.size();
    jaz::plain_array<uint64_t> sh2(l - kmer2 + 1, 0);

    unsigned int v_sz = kmer2 * sizeof(uint64_t);

    for (unsigned int i = 0; i < l - kmer2 + 1; ++i) {
	const char* v = reinterpret_cast<const char*>(sh1.begin() + i);
	sh2[i] = hash(v, v_sz);
    }

    return *std::min_element(sh2.begin(), sh2.end());
} // super_shingle


int main(int argc, char* argv[]) {
    if ((argc < 3) || (argc > 6)) {
	std::cout << "Usage: closet-fasta-preprocess <fasta> <data> [link] [kmer1] [kmer2]" << std::endl;
	return -1;
    }

    const unsigned int SEQ_BUF = 512000;

    // check input
    bool link = false;
    unsigned int kmer1 = 20;
    unsigned int kmer2 = 5;

    if (argc > 3) link = true;

    if (argc > 4) {
	kmer1 = std::atoi(argv[4]);
	if (kmer1 < 3) {
	    std::cerr << error << "incorrect kmer1 value, should be greater than 2" << std::endl;
	    return -1;
	}
    }

    if (argc > 5) {
	kmer2 = std::atoi(argv[5]);
	if (kmer2 < 2) {
	    std::cerr << error << "incorrect kmer2 value, should be greater than 1" << std::endl;
	    return -1;
	}
    }

    // get list of files
    fs::path p(argv[1]);

    if (fs::exists(p) == false) {
	std::cerr << error << "unable to read " << argv[1] << std::endl;
	return -1;
    }

    std::vector<fs::path> files;

    if (fs::is_directory(p) == true) std::copy(fs::directory_iterator(p), fs::directory_iterator(), std::back_inserter(files));
    else files.push_back(p);

    if (files.empty() == true) {
	std::cerr << error << "unable to find input files" << std::endl;
	return -1;
    }

    // open output files in advance to avoid surprises
    std::string out_name = argv[2];

    std::ofstream fmap((out_name + ".map").c_str());
    if (!fmap) {
	std::cerr << error << "unable to create " << out_name + ".map" << std::endl;
	return -1;
    }

    std::ofstream fseq((out_name + ".seq").c_str());
    if (!fseq) {
	std::cerr << error << "unable to create " << out_name + ".seq" << std::endl;
	return -1;
    }

    std::ofstream fidx((out_name + ".idx").c_str(), std::ios_base::binary);
    if (!fidx) {
	std::cerr << error << "unable to create " << out_name + ".idx" << std::endl;
	return -1;
    }


    // *** SUPER-SHINGLE SEQUENCES ***
    unsigned int l = 0;

    std::vector<fasta_type> sf;
    std::vector<SDescriptor> sd;

    // go over all files
    for (unsigned int i = 0; i < files.size(); ++i) {
	if (fs::is_regular_file(files[i]) == false) continue;

	bIO::FASTA_input fasta(files[i].string());

	if (!fasta) {
	    std::cout << "ignoring " << files[i].string() << std::endl;
	    continue;
	} else std::cout << "reading " << files[i].string() << std::endl;

	// read FASTA files
	while (++fasta) {
	    const fasta_type& v = *fasta;

	    if ((v.second.size() < kmer1) || (v.second.size() - kmer1 + 1 < kmer2)) {
		std::cout << "skipping sequence " << v.first << std::endl;
		continue;
	    }

	    std::string s = boost::lexical_cast<std::string>(v.second);
	    clean_sequence(s);

	    sf.push_back(v);
	    sf.back().second = s;

	    // if buffer is full process in parallel
	    if (sf.size() > SEQ_BUF) {
		if (link == true) std::cout << "processing " << sf.size() << " sequences" << std::endl;

		unsigned int p = sd.size();
		sd.resize(p + sf.size());

		l = sf.size();

#pragma omp parallel for shared(sd)
		for (unsigned int j = 0; j < l; ++j) {
		    sd[p + j].name = sf[j].first;
		    sd[p + j].len = sf[j].second.size();
		    if (link == true) sd[p + j].sh = super_shingle(sf[j].second, kmer1, kmer2);
		} // for j

		sf.clear();
	    } // if sf

	} // while

    } // for i

    // take care of unprocessed buffer
    if (sf.empty() == false) {
	if (link == true) std::cout << "processing " << sf.size() << " sequences" << std::endl;

	unsigned int p = sd.size();
	sd.resize(p + sf.size());

	l = sf.size();

#pragma omp parallel for shared(sd)
	for (unsigned int j = 0; j < l; ++j) {
	    sd[p + j].name = sf[j].first;
	    sd[p + j].len = sf[j].second.size();
	    if (link == true) sd[p + j].sh = super_shingle(sf[j].second, kmer1, kmer2);
	} // for j

    } // if sf

    std::vector<fasta_type>().swap(sf);


    // *** CLUSTER SEQUENCES ***

    if (sd.empty() == true) {
	std::cerr << error << "unable to find input sequences" << std::endl;
	return -1;
    }

    l = sd.size();

    jaz::plain_array<unsigned int> uf(sd.size());
    for (unsigned int i = 0; i < l; ++i) uf[i] = i;

    typedef std::vector<unsigned int> intseq_type;
    boost::unordered_map<uint64_t, intseq_type>::iterator iter;

    if (link == true) {
	boost::unordered_map<uint64_t, intseq_type> h2rid;
	for (unsigned int i = 0; i < l; ++i) h2rid[sd[i].sh].push_back(i);

	iter = h2rid.begin();

	for (; iter != h2rid.end(); ++iter) {
	    l = iter->second.size();
	    for (unsigned int i = 1; i < l; ++i) {
		jaz::set_union(iter->second[0], iter->second[i], uf.begin());
	    }
	} // for iter
    } // if link

    l = uf.size();
    boost::unordered_map<std::size_t, intseq_type> CLS;

    for (unsigned int i = 0; i < l; ++i) CLS[jaz::set_find(i, &uf[0])].push_back(i);

    // find the longest sequence for a given cluster
    // put it as first element
    for (iter = CLS.begin() ; iter != CLS.end(); ++iter) {
	unsigned int mel = 0;
	for (unsigned int i = 1; i < iter->second.size(); ++i) {
	    if (sd[iter->second[mel]].len < sd[iter->second[i]].len) mel = i;
	}
	std::swap(iter->second[mel], iter->second[0]);
    } // for iter


    // *** OUTPUT SEQUENCES ***

    // write sequence map
    boost::unordered_map<std::string, unsigned int> seq2id;
    for (unsigned int i = 0; i < sd.size(); ++i) seq2id[sd[i].name] = i;

    for (iter = CLS.begin() ; iter != CLS.end(); ++iter) {
	fmap << iter->second[0] << " " << iter->second.size() << std::endl;
	for (unsigned int i = 0; i < iter->second.size(); ++i) {
	    fmap << sd[iter->second[i]].name << std::endl;
	}
    }

    fmap.close();

    // write index and sequences
    unsigned int offset = 0;

    SequenceIndex sidx;
    SequenceCodec SQ;

    for (unsigned int i = 0; i < files.size(); ++i) {
	if (fs::is_regular_file(files[i]) == false) continue;

	bIO::FASTA_input fasta(files[i].string());

	if (!fasta) {
	    std::cout << "ignoring " << files[i].string() << std::endl;
	    continue;
	} else std::cout << "converting " << files[i].string() << std::endl;

	while (++fasta) {
	    const fasta_type& v = *fasta;

	    if ((v.second.size() < kmer1) || (v.second.size() - kmer1 + 1 < kmer2)) {
		std::cout << "skipping sequence " << v.first << std::endl;
		continue;
	    }

	    sidx.id = seq2id[v.first];

	    if (CLS[uf[sidx.id]][0] == sidx.id) {
		std::string s = boost::lexical_cast<std::string>(v.second);
		clean_sequence(s);

		s = SQ.code(s);

		fseq << nascii85_encode(sidx.id) << '\t' << s << '\n';

		// this must be length of uncompressed sequence
		// sidx.len = s.size();
		sidx.len = v.second.size();

		// we are not using offset anymore
		offset += boost::lexical_cast<std::string>(sidx.id).size() + 1;
		sidx.off = offset;
		offset += s.size() + 1;

		fidx.write(reinterpret_cast<char*>(&sidx), sizeof(sidx));
	    } // if
	} // while
    } // for i

    fidx.close();
    fseq.close();

    std::cout << "done" << std::endl;

    return 0;
} // main
